import os
import yaml
import copy
import pkg_resources

class Repository:
  
  def __init__(self, path):
    self.path = path

  def load_config(self):
      '''return a dictionary with the whole configuration for this repository'''
      config = self.load_config_file()
      config['profiles'] = self.load_profiles()
      config['modules'] = self.load_modules(config)
      self.normalize_profiles(config['profiles'])
      return config

  def get_module_manifests(self):
      mypath = os.path.join(self.path, 'modules')
      if os.path.exists(mypath):
        return [os.path.join(mypath, f) for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f)) and not f.startswith('.')]
      else:
        return []

  def get_profile_manifests(self):
      mypath = os.path.join(self.path, 'profiles')
      if os.path.exists(mypath):
        return [os.path.join(mypath, f) for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f)) and not f.startswith('.')]
      else:
        return []

  def get_scripts_location(self):
      return os.path.join(self.path, 'scripts')

  def load_config_file(self):
      config = {}
      config_file = os.path.join(self.path, 'config.yaml')
      if os.path.exists(config_file):
        with (open(config_file)) as f:
            config = yaml.load(f, Loader=yaml.Loader)
      return config

  def load_profiles(self):
      profiles = []
      for manifest in self.get_profile_manifests():
          with (open(manifest)) as f:
              profile = yaml.load(f, Loader=yaml.Loader)
              profiles.append(profile)

      return profiles
          
  def load_modules(self, config):
      modules = {}
      for manifest in self.get_module_manifests():
          with (open(manifest)) as f:
              module = yaml.load(f, Loader=yaml.Loader)
              modules[module['name']] = module
              if 'parameters' not in module['analysis']:
                  module['analysis']['parameters'] = {}
              if 'parameters' not in module['converter']:
                  module['converter']['parameters'] = {}
              if module['analysis']['parameters'] is None:
                  module['analysis']['parameters'] = {}
              if module['converter']['parameters'] is None:
                  module['converter']['parameters'] = {}
              if not 'container' in module['analysis'] or module['analysis']['container'] is None:
                  module['analysis']['container'] = {"docker":None, "singularity":None}
              elif not 'docker' in module['analysis']['container']:
                  module['analysis']['container']['docker'] = None
              elif not 'singularity' in module['analysis']['container']:
                  module['analysis']['container']['singularity'] = None
              
              if 'database' in module['analysis']['parameters'] and module['analysis']['parameters']['database']:
                  database = module['analysis']['parameters']['database']
                  if not database.startswith("/") and 'databases' in config and 'localpath' in config['databases']:
                      # relative path, resolve it to the database directory
                      module['analysis']['parameters']['database'] = os.path.join(config['databases']['localpath'], database)

              module['analysis']['script'] = os.path.abspath(os.path.join(self.get_scripts_location(), module['analysis']['script']))
              module['converter']['script'] = os.path.abspath(os.path.join(self.get_scripts_location(), module['converter']['script']))
      return modules

  def normalize_profiles(self, profiles):
      """take care that all keys of the profile map exist"""
      for profile in profiles:
          if 'modules' not in profile:
              profile['modules'] = []
          for module in profile['modules']:
              if profile['modules'][module] is None:
                  profile['modules'][module] = {}

class Config:

  def __init__(self, repository_paths):
    self.repo_paths = repository_paths

  def merge_config(self, source, target):
    import jsonmerge
    # Quick hack: replacement instead of merge
    # TODO do an actual merge instead of just a replacement

    schema = { 'properties': {'profiles':{'mergeStrategy':'append'}} }
    return jsonmerge.merge(target, source, schema)

  def load_config(self):
      config = {}
      config['app'] = {}
      for repo_path in self.repo_paths:
        repo = Repository(repo_path)
        repo_config = repo.load_config()
        config = self.merge_config(repo_config, config)
        # quick hack
        # use the last script directory
        # TODO replace with list
        config['app']['script_path'] = repo.get_scripts_location()

      config['install_path'] = pkg_resources.resource_filename('psot', '')
      config['helpers_path'] = pkg_resources.resource_filename('psot', 'helpers')
      config['repository_paths'] = self.repo_paths

      # check if running inside a virtual environment
      if 'VIRTUAL_ENV' in os.environ:
        config['venv'] = os.environ['VIRTUAL_ENV']

      self.merge_modules_in_profiles(config)

      # check that tools is a dictionary
      if 'tools' not in config or config['tools'] is None:
        config['tools'] = {}
      return config

  def merge_modules_in_profiles(self, config):
      """Merge the module configuration and the profile configuration. Profile has precedence over module."""
      profiles = config['profiles']
      modules = config['modules']
      for profile in profiles:
          resolved_modules = []
          for module_name in profile['modules']:
              profile_module_config = profile['modules'][module_name]
              resolved_module = modules[module_name]
              copied_module = copy.deepcopy(resolved_module)
              for c in profile_module_config:
                  copied_module['analysis']['parameters'][c] = profile_module_config[c]
              resolved_modules.append(copied_module)
          profile['modules'] = resolved_modules


def load_config(arguments=None, repositories=None):

    global container_run
    container_run = {}
    if arguments:
        container_run['docker?'] = arguments.docker
        container_run['singularity?'] = arguments.singularity
    else:
        container_run['docker?'] = False
        container_run['singularity?'] = False
    
    # always use the installed repository as the first repo
    all_repositories = [pkg_resources.resource_filename('psot', '')]

    # check for repositories via environment variable
    if 'PSOT_REPOSITORIES' in os.environ:
      variable = os.environ['PSOT_REPOSITORIES']
      env_repos = variable.split(':')
      # convert all paths to absolute paths
      env_repos = [os.path.abspath(x) for x in env_repos]
      os.environ['PSOT_REPOSITORIES'] = ":".join(env_repos)
      all_repositories.extend(env_repos)

    if repositories:
      all_repositories.extend(repositories)

    return Config(all_repositories).load_config()

if __name__ == "__main__":
    """prints the configuration"""
    from pprint import pformat
    config = Config([pkg_resources.resource_filename('psot', '')])

    print('Install location: ' + str(config.repo_paths))
    print('Aggregated config:\n' + pformat(config.load_config()) + '\n')

